#!/usr/bin/env python3

import sys, os
QMCCHEM_PATH=os.environ["QMCCHEM_PATH"]
sys.path.insert(0,QMCCHEM_PATH+"/EZFIO/Python/")
from ezfio import ezfio
from datetime import datetime
import time
import numpy as np
import subprocess
import atexit
import scipy as sp
import scipy.optimize
from math import sqrt

#------------------------------------------------------------------------------
def make_atom_map():
    labels    = {}
    dimension = 0
    # i: label   of nuclei
    # k: counter of nuclei
    for i,k in enumerate(ezfio.nuclei_nucl_label):
        if k in labels:                
            labels[k].append(i)            
        else:
            labels[k] = [dimension, i]
            dimension += 1
    atom_map = [[] for i in range(dimension)]
    for atom in labels.keys():
        l = labels[atom]
        atom_map[l[0]] = l[1:]
    return atom_map
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_params_pen():
    d = ezfio.jastrow_jast_pen
    return np.array([d[m[0]] for m in atom_map])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def set_params_pen(x):
    #x = np.abs(x)
    y = list(ezfio.jastrow_jast_pen)
    for i,m in enumerate(atom_map):
        for j in m:
            y[j] = x[i]
    ezfio.set_jastrow_jast_pen(y)
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_params_b():
    b = ezfio.get_jastrow_jast_b_up_up()
    return b
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def set_params_b(b):
    ezfio.set_jastrow_jast_b_up_up(b)
    ezfio.set_jastrow_jast_b_up_dn(b)
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_energy():
    buffer = subprocess.check_output( 
            ['qmcchem', 'result', '-e', 'e_loc', EZFIO_file], encoding='UTF-8')
    if buffer.strip() != "":
        buffer = buffer.splitlines()[-1]
        _, energy, error = [float(x) for x in buffer.split()]
        return energy, error
    else:
        return None, None
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_variance():
    buffer = subprocess.check_output( 
       ['qmcchem', 'result', '-e', 'e_loc_qmcvar', EZFIO_file], encoding='UTF-8')
    if buffer.strip() != "":
        buffer = buffer.splitlines()[-1]
        _, variance, error = [float(x) for x in buffer.split()]
        return variance, error
    else:
        return None, None
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def set_vmc_params(block_time,total_time):
    #subprocess.check_output(['qmcchem', 'edit', '-c', '-j', 'Simple',
    #                         '-m', 'VMC',
    #                         '-l', str(block_time),
    #                         '--time-step=0.3',
    #                         '--stop-time=36000',
    #                         '--norm=1.e-5',
    #                         '-w', '10',
    #                         EZFIO_file])
    subprocess.check_output(['qmcchem', 'edit', '-c', '-j', 'Simple',
                             '-t', str(total_time),
                             '-l', str(block_time), EZFIO_file])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def run_qmc():
    return subprocess.check_output(['qmcchem', 'run', EZFIO_file])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def stop_qmc():
    subprocess.check_output(['qmcchem', 'stop', EZFIO_file])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def f(x):
    # !!!
    print('f on:')
    print('nuc param Jast = {}'.format(x[:-1]))
    print('b   param Jast = {}'.format(x[-1]))
    # !!!
    global i_fev
    i_fev = i_fev + 1
    # !!!
    global memo_energy
    # !!!
    print ("x = %s"%str(x))
    h = str(x)
    if h in memo_energy:
        return memo_energy[h]
    # !!!
    set_params_pen(x[:-1])
    set_params_b(x[-1])
    set_params_pen(x)
    set_vmc_params(block_time_f, total_time_f)
    # !!!
    pid = os.fork()
    if pid == 0:
        run_qmc()
        os._exit(os.EX_OK)
    else:
        atexit.register(stop_qmc)
        time.sleep(3.*block_time_f/4.)

        local_thresh = thresh
        err          = thresh+1.
        ii           = 1
        ii_max       = int(total_time_f/block_time_f) - 1
        while( (err > local_thresh) and (ii<ii_max) ):
            time.sleep(block_time_f)
            e, e_err        = get_energy()
            variance, v_err = get_variance()
            if e is None or variance is None:
                continue
            energy = e + variance # minimise energy + variance
            err    = sqrt(e_err*e_err+v_err*v_err)
            print(" energy: %f  %f "%(e, e_err))
            print(" varian: %f  %f "%(variance, v_err))
            print(" additi: %f  %f "%(energy, err))
            print(" ")
            if (energy-2.*err) > memo_energy['fmin']+thresh:
               local_thresh = 10.*thresh
            elif (energy+2.*err) < memo_energy['fmin']-thresh:
               local_thresh = 10.*thresh

            ii = ii + 1
            # Check if PID is still running
            try:
                os.kill(pid,0)
            except OSError:
                print("---")
                break
        stop_qmc()
    # !!!
    os.wait()
    memo_energy[h]      = energy + err
    memo_energy['fmin'] = min(energy, memo_energy['fmin'])
    # !!!
    return energy
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------







if __name__ == '__main__':
    t0 = time.time()
    # !!!
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ #
    EZFIO_file = "/home/ammar/qp2/src/svdwf/h2o_optJast"
    # nuclear energy
    E_toadd = 9.194966082434476
    # PARAMETERS
    thresh = 1.e-2
    # maximum allowed number of function evaluations
    N_fev = 4
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ #
    # !!!
    ezfio.set_file(EZFIO_file)
    print("Today's date:", datetime.now() )
    print("EZFIO file = {}".format(EZFIO_file))
    # !!!
    # map nuclei to a list
    # for H2O this will give: atom_map = [[0], [1, 2]]
    atom_map = make_atom_map()
    n_nucpar = len(atom_map)  # nb of nclear parameters
    # !!!
    # x = get_params_pen()
    x = [1.25816521, 0.23766714]
    print('initial pen: {}'.format(x))
    b_par = get_params_b()
    print('initial b: {}'.format(b_par))
    x.append(b_par)
    # !!!
    i_fev = 0
    bnds  = [(0.001, 9.99) for _ in range(n_nucpar)]
    memo_energy = {'fmin': 100000000.}
    opt = sp.optimize.minimize(f, x, method="Powell", bounds=bnds
                                , options= {'disp':True, 
                                            'ftol':thresh, 
                                            'xtol':0.2, 
                                            'maxfev':N_fev} )
    print("x = "+str(opt))
    set_params_pen(opt['x'])
    print('number of function evaluations = {}'.format(i_fev))
    # !!!
    print('memo_energy: {}'.format(memo_energy))
    # !!!
    print("end after {:.3f} minutes".format((time.time()-t0)/60.) )
    # !!!
