#!/usr/bin/env python3

import sys, os
QMCCHEM_PATH=os.environ["QMCCHEM_PATH"]
sys.path.insert(0,QMCCHEM_PATH+"/EZFIO/Python/")
from ezfio import ezfio
from datetime import datetime
import time
import numpy as np
import subprocess
import atexit
import scipy as sp
import scipy.optimize
from math import sqrt

#------------------------------------------------------------------------------
def make_atom_map():
    labels    = {}
    dimension = 0
    # i: label   of nuclei
    # k: counter of nuclei
    for i,k in enumerate(ezfio.nuclei_nucl_label):
        if k in labels:                
            labels[k].append(i)            
        else:
            labels[k] = [dimension, i]
            dimension += 1
    atom_map = [[] for i in range(dimension)]
    for atom in labels.keys():
        l = labels[atom]
        atom_map[l[0]] = l[1:]
    return atom_map
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_params_pen():
    d = ezfio.jastrow_jast_pen
    return np.array([d[m[0]] for m in atom_map])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def set_params_pen(x):
    y = list(ezfio.jastrow_jast_pen)
    for i,m in enumerate(atom_map):
        for j in m:
            y[j] = x[i]
    ezfio.set_jastrow_jast_pen(y)
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_params_b():
    b = ezfio.get_jastrow_jast_b_up_up()
    return b
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def set_params_b(b):
    ezfio.set_jastrow_jast_b_up_up(b)
    ezfio.set_jastrow_jast_b_up_dn(b)
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_energy():
    buffer = subprocess.check_output( 
            ['qmcchem', 'result', '-e', 'e_loc', EZFIO_file], encoding='UTF-8')
    buffer = buffer.splitlines()[-1]
    _, energy, error = [float(x) for x in buffer.split()]
    return energy, error
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_variance():
    buffer = subprocess.check_output( 
       ['qmcchem', 'result', '-e', 'e_loc_qmcvar', EZFIO_file], encoding='UTF-8')
    buffer = buffer.splitlines()[-1]
    _, variance, error = [float(x) for x in buffer.split()]
    return variance, error
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_energy_deriv_nucPar():
    # !!!
    buffer = subprocess.check_output(
             ['qmcchem', 'result', EZFIO_file], encoding='UTF-8')
    end    = len(buffer)
    # !!!
    deriv_nucPar1 = []
    beg           = buffer.find('E_deriv_nucpar_loc1 : [ ') 
    beg           = beg +  len( 'E_deriv_nucpar_loc1 : [ ' )
    der_buf       = buffer[beg:end]
    der_buf       = der_buf.split( '\n' )
    for iline in range(1, n_nucpar+1):
        line = der_buf[iline].split()
        deriv_nucPar1.append( float(line[2]) )
    # !!!
    deriv_nucPar2 = []
    beg           = buffer.find('E_deriv_nucpar_loc2 : [ ') 
    beg           = beg +  len( 'E_deriv_nucpar_loc2 : [ ' )
    der_buf       = buffer[beg:end]
    der_buf       = der_buf.split( '\n' )
    for iline in range(1, n_nucpar+1):
        line          = der_buf[iline].split()
        deriv_nucPar2.append( float(line[2]) )
    # !!!
    deriv_nucPar = []
    e, _ = get_energy()
    for i in range(n_nucpar):
        tmp = 2. * ( deriv_nucPar1[i] - e * deriv_nucPar2[i] )
        deriv_nucPar.append( tmp )
    # !!!
    return deriv_nucPar
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def get_energy_deriv_bPar():
    # !!!
    buffer = subprocess.check_output(
             ['qmcchem', 'result', EZFIO_file], encoding='UTF-8')
    end    = len(buffer)
    # !!!
    beg         = buffer.find('E_deriv_bpar_loc1 : ') 
    beg         = beg +  len( 'E_deriv_bpar_loc1 : ' )
    der_buf     = buffer[beg:end]
    der_buf     = der_buf.split( '\n' )
    der_buf     = der_buf[0].split()
    deriv_bPar1 = float(der_buf[0])
    # !!!
    beg         = buffer.find('E_deriv_bpar_loc2 : ') 
    beg         = beg +  len( 'E_deriv_bpar_loc2 : ' )
    der_buf     = buffer[beg:end]
    der_buf     = der_buf.split( '\n' )
    der_buf     = der_buf[0].split()
    deriv_bPar2 = float(der_buf[0])
    # !!!
    e, _       = get_energy()
    deriv_bPar = deriv_bPar1 - e * deriv_bPar2
    # !!!
    return deriv_bPar
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def set_vmc_params(block_time,total_time):
    subprocess.check_output(['qmcchem', 'edit', '-c', '-j', 'Simple',
                             '-t', str(total_time),
                             '-l', str(block_time), EZFIO_file])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def run_qmc():
    return subprocess.check_output(['qmcchem', 'run', EZFIO_file])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def stop_qmc():
    subprocess.check_output(['qmcchem', 'stop', EZFIO_file])
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def f(x):
    # !!!
    print('f on:')
    print('nuc param Jast = {}'.format(x[:-1]))
    print('b   param Jast = {}'.format(x[-1]))
    # !!!
    global i_fev
    i_fev = i_fev + 1
    # !!!
    set_params_pen(x[:-1])
    set_params_b(x[-1])
    block_time_f = 20
    total_time_f = 40
    set_vmc_params(block_time_f, total_time_f)
    # !!!
    err    = 10.
    ii     = 0
    ii_max = 5
    while( (err > thresh) and (ii<ii_max) ):
        ii = ii + 1
        run_qmc()
        energy, err = get_energy()
        print(" energy: %f  %f "%(energy, err))
    print(" ")
    # !!!
    global memo_energy
    h                   = str(x)
    memo_energy[h]      = energy + err
    memo_energy['fmin'] = min(energy, memo_energy['fmin'])
    # !!!
    return energy
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------
def fprime(x):
    # !!!
    # used with CG, BFGS, Newton-CG, L-BFGS-B, TNC, SLSQP, dogleg, trust-ncg, 
    # trust-krylov, trust-exact, trust-constr
    # !!!
    print('derivative on:')
    print('nuc param Jast = {}'.format(x[:-1]))
    print('b   param Jast = {}'.format(x[-1]))
    # !!!
    set_params_pen(x[:-1])
    set_params_b(x[-1])
    block_time_fp = 20
    total_time_fp = 30
    set_vmc_params(block_time_fp, total_time_fp)
    run_qmc()
    # !!!
    fp = get_energy_deriv_nucPar()
    print('nuc param Jast = {}'.format(fp))
    fp_bpar = get_energy_deriv_bPar()
    print('b   param Jast = {}'.format(fp_bpar))
    fp.append(fp_bpar)
    print(" ")
    # !!!
    return fp
#------------------------------------------------------------------------------
##
###
##
#------------------------------------------------------------------------------







if __name__ == '__main__':
    # !!!
    t0 = time.time()
    # !!!
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ #
    EZFIO_file = "/home/ammar/qp2/src/svdwf/h2o_optJast"
    # nuclear energy
    E_toadd = 9.194966082434476
    # PARAMETERS
    thresh = 1.e-2
    # maximum allowed number of function evaluations
    N_fev = 5
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ #
    # !!!
    ezfio.set_file(EZFIO_file)
    print("Today's date:", datetime.now() )
    print("EZFIO file = {}".format(EZFIO_file))
    # !!!
    # map nuclei to a list
    # for H2O this will give: atom_map = [[0], [1, 2]]
    atom_map = make_atom_map()
    n_nucpar = len(atom_map)  # nb of nclear parameters
    # !!!
    # x = get_params_pen()
    x = [1.25816521, 0.35]
    print('initial pen: {}'.format(x))
    b_par = get_params_b()
    print('initial b: {}'.format(b_par))
    x.append(b_par)
    # !!!
    #i_fev = 0
    #bnds  = [(0.001, 9.99) for _ in range(n_nucpar)]
    #memo_energy = {'fmin': 100000000.}
    #opt = sp.optimize.minimize(f, x, method="Powell", bounds=bnds
     #                           , options= {'disp':True, 
     #                                       'ftol':thresh, 
     #                                       'xtol':0.2, 
     #                                       'maxfev':N_fev} )
    #print("x = "+str(opt))
    #set_params_pen(opt['x'])
    #print('number of function evaluations = {}'.format(i_fev))
    # !!!
    #print('memo_energy: {}'.format(memo_energy))
    # !!!
    #stop_qmc()
    print("end after {:.3f} minutes".format((time.time()-t0)/60.) )
    # !!!
